@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2016-2020, OARC, Inc.
+ * Copyright (c) 2016-2025 OARC, Inc.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -40,6 +40,9 @@
 #include "log.h"
 #include "tcpstate.h"
 #include "network.h"
+#include "dumper.h"
+
+#include <ldns/ldns.h>
 
 /*
  * OpenBSD and Debian Stretch i386 need file local functions for export
@@ -111,6 +114,16 @@ void set_iaddr(iaddr* from, iaddr* to)
     }
 }
 
+long _get_pcap_thread_ftell(void)
+{
+    return pcap_thread_ftell(&pcap_thread);
+}
+
+size_t _get_pkthdr_caplen(void)
+{
+    return pkthdr_caplen;
+}
+
 #ifdef __linux__
 extern char* strptime(const char*, const char*, struct tm*);
 #endif
@@ -150,7 +163,7 @@ void help_1(void)
     fprintf(stderr, "%s: version %s\n\n", ProgramName, PACKAGE_VERSION);
     fprintf(stderr,
         "usage: %s\n"
-        "  [-?VbNpd1g6fTI"
+        "  [-?VbNpd1gfTI"
 #ifdef USE_SECCOMP
         "y"
 #endif
@@ -162,7 +175,8 @@ void help_1(void)
         "  [-t <lim>] [-c <lim>] [-C <lim>]\n"
         "  [-x <pat>]+ [-X <pat>]+\n"
         "  [-B <datetime>] [-E <datetime>]\n"
-        "  [-U <str>] [-P plugin.so <plugin options...>]\n",
+        "  [-U <str>] [-q <num|str>] [-Q <num|str>]\n"
+        "  [-P plugin.so <plugin options...>]\n",
         ProgramName);
 }
 
@@ -182,7 +196,6 @@ void help_2(void)
         "             times to increase debugging\n"
         "  -1         flush output on every packet\n"
         "  -g         dump packets dig-style on stderr\n"
-        "  -6         compensate for PCAP/BPF IPv6 bug\n"
         "  -f         include fragmented packets\n"
         "  -T         include TCP packets (DNS header filters will inspect only the\n"
         "             first DNS header, and the result will apply to all messages\n"
@@ -228,6 +241,8 @@ void help_2(void)
         "  -M         set monitor mode on interfaces\n"
         "  -D         set immediate mode on interfaces\n"
         "  -U <str>   append 'and <str>' to the pcap filter\n"
+        "  -q <num|str> select messages based on QTYPE\n"
+        "  -Q <num|str> filter out messages based on QTYPE\n"
         "  -P <plugin.so> load plugin, any argument after this is sent to the plugin!\n");
 }
 
@@ -262,7 +277,7 @@ void parse_args(int argc, char* argv[])
     vlan_ptr      vlan;
     unsigned      u;
     int           ch;
-    char*         p;
+    char *        p, *match_qtype_arg = 0;
 
     if ((p = strrchr(argv[0], '/')) == NULL)
         ProgramName = argv[0];
@@ -279,13 +294,13 @@ void parse_args(int argc, char* argv[])
     INIT_LIST(myregexes);
     INIT_LIST(plugins);
     while ((ch = getopt(argc, argv,
-                "a:bc:de:fgh:i:k:l:m:o:pr:s:t:u:w:x:yz:"
-                "A:B:C:DE:F:IL:MNP:STU:VW:X:Y:Z:16?"))
+                "a:bc:de:fgh:i:k:l:m:o:pr:s:t:u:w:x:yz:q:"
+                "A:B:C:DE:F:IL:MNP:STU:VW:X:Y:Z:Q:1?"))
            != EOF) {
         switch (ch) {
         case 'o':
             if (option_parse(&options, optarg)) {
-                fprintf(stderr, "%s: unknown or invalid extended option: %s\n", ProgramName, optarg);
+                fprintf(stderr, "%s: unknown or invalid extended -o option: %s\n", ProgramName, optarg);
                 exit(1);
             }
             break;
@@ -307,9 +322,6 @@ void parse_args(int argc, char* argv[])
         case 'g':
             preso = TRUE;
             break;
-        case '6':
-            v6bug = TRUE;
-            break;
         case 'f':
             wantfrags = TRUE;
             break;
@@ -331,7 +343,7 @@ void parse_args(int argc, char* argv[])
             break;
         case 'r':
             if (!EMPTY(mypcaps))
-                usage("-r makes no sense after -i");
+                usage("-r makes no sense after -i, or specified multiple times");
             pcap_offline = calloc(1, sizeof *pcap_offline);
             assert(pcap_offline != NULL);
             INIT_LINK(pcap_offline, link);
@@ -343,7 +355,7 @@ void parse_args(int argc, char* argv[])
         case 'l':
             ul = strtoul(optarg, &p, 0);
             if (*p != '\0' || ul > MAX_VLAN)
-                usage("vlan must be an integer 0..4095");
+                usage("-l vlan must be an integer 0..4095");
             vlan = calloc(1, sizeof *vlan);
             assert(vlan != NULL);
             INIT_LINK(vlan, link);
@@ -359,7 +371,7 @@ void parse_args(int argc, char* argv[])
         case 'L':
             ul = strtoul(optarg, &p, 0);
             if (*p != '\0' || ul > MAX_VLAN)
-                usage("vlan must be an integer 0..4095");
+                usage("-L vlan must be an integer 0..4095");
             vlan = calloc(1, sizeof *vlan);
             assert(vlan != NULL);
             INIT_LINK(vlan, link);
@@ -503,6 +515,8 @@ void parse_args(int argc, char* argv[])
                 options.dump_format = cbor;
             } else if (!strcmp(optarg, "cds")) {
                 options.dump_format = cds;
+            } else if (!strcmp(optarg, "tcpdns")) {
+                options.dump_format = tcpdns;
             } else {
                 usage("invalid output format for -F");
             }
@@ -527,9 +541,7 @@ void parse_args(int argc, char* argv[])
             break;
         case 'x':
         /* FALLTHROUGH */
-        case 'X':
-#if HAVE_NS_INITPARSE && HAVE_NS_PARSERR && HAVE_NS_SPRINTRR
-        {
+        case 'X': {
             int         i;
             myregex_ptr myregex = calloc(1, sizeof *myregex);
             assert(myregex != NULL);
@@ -543,30 +555,19 @@ void parse_args(int argc, char* argv[])
             }
             myregex->not = (ch == 'X');
             APPEND(myregexes, myregex, link);
-        }
-#else
-            /*
-             * -x and -X options require libbind because
-             * the code calls ns_initparse(), ns_parserr(),
-             * and ns_sprintrr()
-             */
-            fprintf(stderr, "%s must be compiled with libbind to use the -x or -X option.\n",
-                ProgramName);
-            exit(1);
-#endif
-        break;
+        } break;
         case 'B': {
             struct tm tm;
             memset(&tm, '\0', sizeof(tm));
             if (NULL == strptime(optarg, "%F %T", &tm))
-                usage("--B arg must have format YYYY-MM-DD HH:MM:SS");
+                usage("-B arg must have format YYYY-MM-DD HH:MM:SS");
             start_time = xtimegm(&tm);
         } break;
         case 'E': {
             struct tm tm;
             memset(&tm, '\0', sizeof(tm));
             if (NULL == strptime(optarg, "%F %T", &tm))
-                usage("--E arg must have format YYYY-MM-DD HH:MM:SS");
+                usage("-E arg must have format YYYY-MM-DD HH:MM:SS");
             stop_time = xtimegm(&tm);
         } break;
         case 'S':
@@ -576,71 +577,74 @@ void parse_args(int argc, char* argv[])
             char*          fn = strdup(optarg);
             char*          t;
             char           sn[256];
-            struct plugin* p = calloc(1, sizeof(*p));
-            assert(p != NULL);
-            INIT_LINK(p, link);
-            t       = strrchr(fn, '/');
-            p->name = strdup(t ? t + 1 : fn);
-            if ((t = strstr(p->name, ".so")))
+            struct plugin* pl = calloc(1, sizeof(*pl));
+            assert(pl != NULL);
+            INIT_LINK(pl, link);
+            t        = strrchr(fn, '/');
+            pl->name = strdup(t ? t + 1 : fn);
+            if ((t = strstr(pl->name, ".so")))
                 *t = 0;
-            p->handle = dlopen(fn, RTLD_NOW);
-            if (!p->handle) {
+            pl->handle = dlopen(fn, RTLD_NOW);
+            if (!pl->handle) {
                 logerr("%s: %s", fn, dlerror());
                 exit(1);
             }
-            snprintf(sn, sizeof(sn), "%s_type", p->name);
-            p->type = dlsym(p->handle, sn);
-            if (p->type) {
-                p->pt = (*p->type)();
-                switch (p->pt) {
+            snprintf(sn, sizeof(sn), "%s_type", pl->name);
+            pl->type = dlsym(pl->handle, sn);
+            if (pl->type) {
+                pl->pt = (*pl->type)();
+                switch (pl->pt) {
                 case plugin_output:
                 case plugin_filter:
                     break;
                 default:
-                    logerr("invalid plugin type for plugin '%s'", p->name);
+                    logerr("invalid plugin type for plugin '%s'", pl->name);
                     exit(1);
                 }
             } else {
-                p->pt = plugin_output;
+                pl->pt = plugin_output;
             }
-            snprintf(sn, sizeof(sn), "%s_start", p->name);
-            p->start = dlsym(p->handle, sn);
-            snprintf(sn, sizeof(sn), "%s_stop", p->name);
-            p->stop = dlsym(p->handle, sn);
-            snprintf(sn, sizeof(sn), "%s_open", p->name);
-            p->open = dlsym(p->handle, sn);
-            snprintf(sn, sizeof(sn), "%s_close", p->name);
-            p->close = dlsym(p->handle, sn);
-            snprintf(sn, sizeof(sn), "%s_output", p->name);
-            p->output = dlsym(p->handle, sn);
-            if (p->pt == plugin_output && !p->output) {
+            snprintf(sn, sizeof(sn), "%s_start", pl->name);
+            pl->start = dlsym(pl->handle, sn);
+            snprintf(sn, sizeof(sn), "%s_stop", pl->name);
+            pl->stop = dlsym(pl->handle, sn);
+            snprintf(sn, sizeof(sn), "%s_open", pl->name);
+            pl->open = dlsym(pl->handle, sn);
+            snprintf(sn, sizeof(sn), "%s_close", pl->name);
+            pl->close = dlsym(pl->handle, sn);
+            snprintf(sn, sizeof(sn), "%s_output", pl->name);
+            pl->output = dlsym(pl->handle, sn);
+            if (pl->pt == plugin_output && !pl->output) {
                 logerr("%s", dlerror());
                 exit(1);
             }
-            snprintf(sn, sizeof(sn), "%s_filter", p->name);
-            p->filter = dlsym(p->handle, sn);
-            if (p->pt == plugin_filter && !p->filter) {
+            snprintf(sn, sizeof(sn), "%s_filter", pl->name);
+            pl->filter = dlsym(pl->handle, sn);
+            if (pl->pt == plugin_filter && !pl->filter) {
                 logerr("%s", dlerror());
                 exit(1);
             }
-            snprintf(sn, sizeof(sn), "%s_usage", p->name);
-            p->usage = dlsym(p->handle, sn);
-            snprintf(sn, sizeof(sn), "%s_extension", p->name);
-            p->extension = dlsym(p->handle, sn);
-            if (p->extension) {
-                (*p->extension)(DNSCAP_EXT_IS_RESPONDER, (void*)is_responder);
-                (*p->extension)(DNSCAP_EXT_IA_STR, (void*)_ia_str);
-                (*p->extension)(DNSCAP_EXT_TCPSTATE_GETCURR, (void*)_tcpstate_getcurr);
-                (*p->extension)(DNSCAP_EXT_TCPSTATE_RESET, (void*)_tcpstate_reset);
-                (*p->extension)(DNSCAP_EXT_SET_IADDR, (void*)set_iaddr);
+            snprintf(sn, sizeof(sn), "%s_usage", pl->name);
+            pl->usage = dlsym(pl->handle, sn);
+            snprintf(sn, sizeof(sn), "%s_extension", pl->name);
+            pl->extension = dlsym(pl->handle, sn);
+            if (pl->extension) {
+                (*pl->extension)(DNSCAP_EXT_IS_RESPONDER, (is_responder_t)is_responder);
+                (*pl->extension)(DNSCAP_EXT_IA_STR, (ia_str_t)_ia_str);
+                (*pl->extension)(DNSCAP_EXT_TCPSTATE_GETCURR, (tcpstate_getcurr_t)_tcpstate_getcurr);
+                (*pl->extension)(DNSCAP_EXT_TCPSTATE_RESET, (tcpstate_reset_t)_tcpstate_reset);
+                (*pl->extension)(DNSCAP_EXT_SET_IADDR, (set_iaddr_t)set_iaddr);
+                (*pl->extension)(DNSCAP_EXT_SET_OUTPUT_PKT, (set_output_pkt_t)set_output_pkt);
+                (*pl->extension)(DNSCAP_EXT_GET_PCAP_THREAD_FTELL, (get_pcap_thread_ftell_t)_get_pcap_thread_ftell);
+                (*pl->extension)(DNSCAP_EXT_GET_PKTHDR_CAPLEN, (get_pkthdr_caplen_t)_get_pkthdr_caplen);
             }
-            snprintf(sn, sizeof(sn), "%s_getopt", p->name);
-            p->getopt = dlsym(p->handle, sn);
-            if (p->getopt)
-                (*p->getopt)(&argc, &argv);
-            APPEND(plugins, p, link);
+            snprintf(sn, sizeof(sn), "%s_getopt", pl->name);
+            pl->getopt = dlsym(pl->handle, sn);
+            if (pl->getopt)
+                (*pl->getopt)(&argc, &argv);
+            APPEND(plugins, pl, link);
             if (dumptrace)
-                fprintf(stderr, "Plugin '%s' loaded\n", p->name);
+                fprintf(stderr, "Plugin '%s' loaded\n", pl->name);
             free(fn);
         } break;
         case 'U':
@@ -653,7 +657,7 @@ void parse_args(int argc, char* argv[])
             use_seccomp = TRUE;
             break;
 #else
-            usage("seccomp-bpf not enabled");
+            usage("-y: seccomp-bpf not enabled");
 #endif
         case 'M':
             monitor_mode = TRUE;
@@ -661,6 +665,36 @@ void parse_args(int argc, char* argv[])
         case 'D':
             immediate_mode = TRUE;
             break;
+        case 'q': {
+            if (nmatch_qtype) {
+                usage("-q and -Q can't be used together");
+            }
+            free(match_qtype_arg); // fix clang scan-build
+            match_qtype_arg = strdup(optarg);
+            match_qtype     = ldns_get_rr_type_by_name(optarg);
+            if (!match_qtype) {
+                ul = strtoul(optarg, &p, 0);
+                if (*p != '\0' || ul < 1U || ul > 65535U)
+                    usage("-q QTYPE must be a valid type or an integer 1..65535");
+                match_qtype = (ldns_rr_type)ul;
+            }
+            break;
+        }
+        case 'Q': {
+            if (match_qtype) {
+                usage("-q and -Q can't be used together");
+            }
+            free(match_qtype_arg); // fix clang scan-build
+            match_qtype_arg = strdup(optarg);
+            nmatch_qtype    = ldns_get_rr_type_by_name(optarg);
+            if (!nmatch_qtype) {
+                ul = strtoul(optarg, &p, 0);
+                if (*p != '\0' || ul < 1U || ul > 65535U)
+                    usage("-Q QTYPE must be a valid type or an integer 1..65535");
+                nmatch_qtype = (ldns_rr_type)ul;
+            }
+            break;
+        }
         case '?':
             if (!optopt || optopt == '?') {
                 help_2();
@@ -674,8 +708,6 @@ void parse_args(int argc, char* argv[])
     }
     assert(msg_wanted != 0U);
     assert(err_wanted != 0U);
-    if (dump_type != nowhere && options.use_layers)
-        usage("use_layers is only compatible with -g so far");
     if (dump_type == nowhere && !preso && EMPTY(plugins))
         usage("without -w or -g, there would be no output");
     if (end_hide != 0U && wantfrags)
@@ -691,7 +723,7 @@ void parse_args(int argc, char* argv[])
 
         fprintf(stderr, "%s: version %s\n", ProgramName, PACKAGE_VERSION);
         fprintf(stderr,
-            "%s: msg %c%c%c, side %c%c, hide %c%c, err %c%c%c%c%c%c%c%c, t %u, c %u, C %zu\n",
+            "%s: msg %c%c%c, side %c%c, hide %c%c, err %c%c%c%c%c%c%c%c, t %u, c %u, C %zu, %sq %s\n",
             ProgramName,
             (msg_wanted & MSG_QUERY) != 0 ? 'Q' : '.',
             (msg_wanted & MSG_UPDATE) != 0 ? 'U' : '.',
@@ -708,7 +740,8 @@ void parse_args(int argc, char* argv[])
             (err_wanted & ERR_NXDOMAIN) != 0 ? 'x' : '.',
             (err_wanted & ERR_NOTIMPL) != 0 ? 'i' : '.',
             (err_wanted & ERR_REFUSED) != 0 ? 'r' : '.',
-            limit_seconds, limit_packets, limit_pcapfilesize);
+            limit_seconds, limit_packets, limit_pcapfilesize,
+            nmatch_qtype ? "!" : "", match_qtype_arg);
         sep = "\tinit";
         for (ep = HEAD(initiators);
              ep != NULL;
@@ -815,12 +848,10 @@ void parse_args(int argc, char* argv[])
     }
 
     if (options.reassemble_tcp_bfbparsedns) {
-#if HAVE_NS_INITPARSE
         if (!options.reassemble_tcp) {
             usage("can't do byte for byte parsing of DNS without reassemble_tcp=yes");
         }
-#else
-        usage("not compiled with libbind, needed for reassemble_tcp_bfbparsedns=yes");
-#endif
     }
+
+    free(match_qtype_arg);
 }
